"""Update iBGP mesh workflow. Adds a new P router to the mesh of PE routers in the network."""

from typing import Any

from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.workflow import StepList, done, init, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import root_validator

from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.router import Router
from gso.services import librenms_client, provisioning_proxy, subscriptions
from gso.services.provisioning_proxy import indifferent_pp_interaction, pp_interaction
from gso.services.subscriptions import get_active_trunks_that_terminate_on_router
from gso.utils.helpers import SNMPVersion


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Show a confirmation window before running the workflow.

    Does not allow for user input, but does run a validation check. The workflow is only allowed to run, if the router
    already is connected by at least one IP trunk.
    """
    subscription = Router.from_subscription(subscription_id)

    class AddBGPSessionForm(FormPage):
        class Config:
            title = f"Add {subscription.router.router_fqdn} to the iBGP mesh?"

        @root_validator
        def router_has_a_trunk(cls, values: dict[str, Any]) -> dict[str, Any]:
            if len(get_active_trunks_that_terminate_on_router(subscription_id)) == 0:
                msg = "Selected router does not terminate any active IP trunks."
                raise ValueError(msg)

            return values

    yield AddBGPSessionForm

    return {"subscription": subscription}


@step("Calculate list of all active PE routers")
def calculate_pe_router_list() -> State:
    """Calculate a list of all active PE routers in the network."""
    all_routers = [
        Router.from_subscription(r["subscription_id"]) for r in subscriptions.get_active_router_subscriptions()
    ]
    all_pe_routers = [router for router in all_routers if router.router.router_role == RouterRole.PE]

    return {"pe_router_list": all_pe_routers}


def _generate_pe_inventory(pe_router_list: list[Router]) -> dict[str, Any]:
    """Generate an Ansible-compatible inventory for executing playbooks. Contains all active PE routers."""
    return {
        "_meta": {
            "vars": {
                router.router.router_fqdn: {
                    "lo4": router.router.router_lo_ipv4_address,
                    "lo6": router.router.router_lo_ipv6_address,
                    "vendor": router.router.vendor,
                }
                for router in pe_router_list
            },
        },
        "all": {
            "hosts": {router.router.router_fqdn: None for router in pe_router_list},
        },
    }


@step("[DRY RUN] Add P router to iBGP mesh")
def add_p_to_mesh_dry(subscription: Router, callback_route: str, pe_router_list: list[Router]) -> State:
    """Perform a dry run of adding the new P router to the PE router mesh."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
    }

    provisioning_proxy.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=_generate_pe_inventory(pe_router_list),
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("[FOR REAL] Add P router to iBGP mesh")
def add_p_to_mesh_real(subscription: Router, callback_route: str, pe_router_list: list[Router]) -> State:
    """Add the P router to the mesh of PE routers."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
    }

    provisioning_proxy.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=_generate_pe_inventory(pe_router_list),
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("[DRY RUN] Add all PE routers to P router iBGP table")
def add_all_pe_to_p_dry(subscription: Router, pe_router_list: list[Router], callback_route: str) -> State:
    """Perform a dry run of adding the list of all PE routers to the new P router."""
    extra_vars = {
        "dry_run": True,
        "pe_router_list": {
            router.router.router_fqdn: {
                "lo4": router.router.router_lo_ipv4_address,
                "lo6": router.router.router_lo_ipv6_address,
                "vendor": router.router.vendor,
            }
            for router in pe_router_list
        },
    }

    inventory = {
        "all": {
            "hosts": {router.router.router_fqdn: None for router in pe_router_list},
        },
    }

    provisioning_proxy.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=inventory,
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("[FOR REAL] Add all PE routers to P router iBGP table")
def add_all_pe_to_p_real(subscription: Router, pe_router_list: list[Router], callback_route: str) -> State:
    """Add the list of all PE routers to the new P router."""
    extra_vars = {
        "dry_run": False,
        "pe_router_list": {
            router.router.router_fqdn: {
                "lo4": router.router.router_lo_ipv4_address,
                "lo6": router.router.router_lo_ipv6_address,
                "vendor": router.router.vendor,
            }
            for router in pe_router_list
        },
    }

    inventory = {
        "all": {
            "hosts": {router.router.router_fqdn: None for router in pe_router_list},
        },
    }

    provisioning_proxy.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=inventory,
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("Verify iBGP session health")
def check_ibgp_session(subscription: Router, callback_route: str) -> State:
    """Run a playbook using the provisioning proxy, to check the health of the new iBGP session."""
    inventory = {
        "all": {
            "hosts": {
                subscription.router.router_fqdn: None,
            },
        },
    }

    provisioning_proxy.execute_playbook(
        playbook_name="check_ibgp.yaml",
        callback_route=callback_route,
        inventory=inventory,
        extra_vars={},
    )

    return {"subscription": subscription}


@step("Add the router to LibreNMS")
def add_device_to_librenms(subscription: Router) -> State:
    """Add the router as a device to LibreNMS."""
    client = librenms_client.LibreNMSClient()
    client.add_device(subscription.router.router_fqdn, SNMPVersion.V2C)

    return {"subscription": subscription}


@step("Update subscription model")
def update_subscription_model(subscription: Router) -> State:
    """Update the database model, such that it should not be reached via :term:`OOB` access anymore."""
    subscription.router.router_access_via_ts = False

    return {"subscription": subscription}


@workflow(
    "Update iBGP mesh",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def update_ibgp_mesh() -> StepList:
    """Update the iBGP mesh with a new P router.

    * Add the new P-router to all other PE-routers in the network, including a dry run.
    * Add all PE-routers to the P-router, including a dry run.
    * Verify that the iBGP session is up.
    * Add the new P-router to LibreNMS.
    * Update the subscription model.
    """
    return (
        init
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> calculate_pe_router_list
        >> pp_interaction(add_p_to_mesh_dry)
        >> pp_interaction(add_p_to_mesh_real)
        >> pp_interaction(add_all_pe_to_p_dry)
        >> pp_interaction(add_all_pe_to_p_real)
        >> indifferent_pp_interaction(check_ibgp_session)
        >> add_device_to_librenms
        >> update_subscription_model
        >> resync
        >> done
    )
